#include "handmade.h"
#include "handmade_tile.cpp"

internal void drawRectangle(GameOffscreenBuffer *buffer, real32 realMinX, real32 realMinY, real32 realMaxX, real32 realMaxY, real32 R, real32 G, real32 B) {
    int32 minX = roundReal32ToInt32(realMinX);
    int32 minY = roundReal32ToInt32(realMinY);
    int32 maxX = roundReal32ToInt32(realMaxX);
    int32 maxY = roundReal32ToInt32(realMaxY);

    uint32 color = (roundReal32ToUInt32(R*255.0f) << 16) |
                   (roundReal32ToUInt32(G*255.0f) <<  8) |
                   (roundReal32ToUInt32(B*255.0f) <<  0);

    if (minX < 0) {
        minX = 0;
    }

    if (minY < 0) {
        minY = 0;
    }

    if (maxX > buffer->width) {
        maxX = buffer->width;
    }

    if (maxY > buffer->height) {
        maxY = buffer->height;
    }

    uint8 *endOfBuffer = (uint8 *)buffer->memory + buffer->pitch*buffer->height;
    uint8 *row = ((uint8 *)buffer->memory + minX*buffer->bytesPerPixel + minY*buffer->pitch);
    for (int y = minY; y < maxY; y++) {
        uint32 *pixel = (uint32 *)row;
        for (int x = minX; x < maxX; x++) {
            *pixel++ = color;
        }
        row += buffer->pitch;
    }
}

internal void outputSound(GameSoundOutputBuffer *soundBuffer, GameState *state) {
    int16 toneVolume = 3000;
    int16 *sampleOut = soundBuffer->samples;
    for (int sampleIndex = 0; sampleIndex < soundBuffer->sampleCount; sampleIndex++) {
        int16 sampleValue = 0;
        *sampleOut++ = sampleValue;
        *sampleOut++ = sampleValue;
    }
}

internal void initialiseArena(MemoryArena *arena, memory_index size, uint8 *base) {
    arena->size = size;
    arena->base = base;
    arena->used = 0;
}

internal void *_pushSize(MemoryArena *arena, memory_index size) {
    Assert((arena->used + size) <= arena->size);
    void *result = (void *)(arena->base + arena->used);
    arena->used += size;
    return result;
}
#define pushStruct(arena, type) (type *)_pushSize(arena, sizeof(type))
#define pushArray(arena, count, type) (type *)_pushSize(arena, (count)*sizeof(type))

extern "C" GAME_UPDATE_AND_RENDER(gameUpdateAndRender) {
    Assert(sizeof(GameState) <= memory->permanentStorageSize);

    GameState *state = (GameState*)memory->permanentStorage;

    const int WORLD_WIDTH = 2;
    const int WORLD_HEIGHT = 2;

    if (!memory->isInitialised) {
        state->playerPos.absTileX = 3;
        state->playerPos.absTileY = 3;
        state->playerPos.x = 5.0f;
        state->playerPos.y = 5.0f;

        initialiseArena(
            &state->worldArena, 
            memory->permanentStorageSize - sizeof(GameState), 
            (uint8*)memory->permanentStorage + sizeof(GameState)
        );

        state->world = pushStruct(&state->worldArena, World);
        World *world = state->world;

        world->tileMap = pushStruct(&state->worldArena, TileMap);
        TileMap *tileMap = state->world->tileMap;

        tileMap->chunkShift = 8;
        tileMap->chunkMask = 0xFF;
        tileMap->chunkDim = 256;
        tileMap->tileChunkCountX = 4;
        tileMap->tileChunkCountY = 4;
        tileMap->tileSideInMeters = 1.4f;
        tileMap->tileSideInPixels = 60;
        tileMap->metersToPixels = tileMap->tileSideInPixels / tileMap->tileSideInMeters;

        tileMap->tileChunks = pushArray(&state->worldArena, tileMap->tileChunkCountX*tileMap->tileChunkCountY, TileChunk);

        for (uint32 y = 0; y < tileMap->tileChunkCountY; y++) {
            for (uint32 x = 0; x < tileMap->tileChunkCountX; x++) {
                tileMap->tileChunks[y*tileMap->tileChunkCountX + x].tiles = pushArray(&state->worldArena, tileMap->chunkDim*tileMap->chunkDim, uint32);
            }
        }


        uint32 tilesPerWidth = 17;
        uint32 tilesPerHeight = 9;

        for (uint32 screenY = 0; screenY < 32; screenY++) {
            for (uint32 screenX = 0; screenX < 32; screenX++) {
                for (uint32 tileY = 0; tileY < tilesPerHeight; tileY++) {
                    for (uint32 tileX = 0; tileX < tilesPerWidth; tileX++) {
                        uint32 absTileX = screenX*tilesPerWidth + tileX;
                        uint32 absTileY = screenY*tilesPerHeight + tileY;
                        setTileMapValue(&state->worldArena, tileMap, absTileX, absTileY, (tileX == tileY && (tileX % 5 == 0)));
                    }
                }
            }
        }

        memory->isInitialised = true;
    }

    real32 screenCenterX = 0.5f*(real32)videoBuf->width;
    real32 screenCenterY = 0.5f*(real32)videoBuf->height;

//    uint32 tiles[TILEMAP_WIDTH][TILEMAP_HEIGHT] = {
//        {1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 1, 1},
//        {1, 0, 0, 0,  0, 1, 1, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 0, 1, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  1, 0, 0, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
//        {1, 1, 1, 1,  1, 1, 1, 0,  0, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 0,  0, 1, 1, 1,  1, 1, 1, 1},
//        {1, 1, 1, 1,  1, 1, 1, 0,  0, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 0,  0, 1, 1, 1,  1, 1, 1, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 1, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 0, 1, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  1, 0, 0, 1},
//        {1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 1, 1, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 1, 0, 1},
//        {1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1, 1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 1, 1},
//    };

    TileMap *tilemap = state->world->tileMap;

    // player
    real32 playerR = 1.0f;
    real32 playerG = 1.0f;
    real32 playerB = 0.5f;
    real32 playerWidth = 0.7f;
    real32 playerHeight = 1.4f;

    for (int controllerIndex = 0; controllerIndex < ArrayCount(input->controllers); controllerIndex++) {
        GameControllerInput *controller = &input->controllers[controllerIndex];

        real32 dPlayerX = 0.0f;
        real32 dPlayerY = 0.0f;

        if (controller->stickUp.endedDown) {
            dPlayerY = 1.0f;
        }
        if (controller->stickDown.endedDown) {
            dPlayerY = -1.0f;
        }
        if (controller->stickLeft.endedDown) {
            dPlayerX = -1.0f;
        }
        if (controller->stickRight.endedDown) {
            dPlayerX = 1.0f;
        }

        dPlayerX *= 6.0f;
        dPlayerY *= 6.0f;

        TileMapPosition newPlayerPos = state->playerPos;
        newPlayerPos.x += dPlayerX * input->dtForFrame;
        newPlayerPos.y += dPlayerY * input->dtForFrame;
        newPlayerPos = recanonicalisePosition(tilemap, newPlayerPos);

        TileMapPosition playerBottomLeft = newPlayerPos;
        playerBottomLeft.x -= 0.5f * playerWidth;
        playerBottomLeft = recanonicalisePosition(tilemap, playerBottomLeft);

        TileMapPosition playerTopLeft = newPlayerPos;
        playerTopLeft.x -= 0.5f * playerWidth;
        playerTopLeft.y += playerWidth;
        playerTopLeft = recanonicalisePosition(tilemap, playerTopLeft);

        TileMapPosition playerBottomRight = newPlayerPos;
        playerBottomRight.x += 0.5f * playerWidth;
        playerBottomRight = recanonicalisePosition(tilemap, playerBottomRight);

        TileMapPosition playerTopRight = newPlayerPos;
        playerTopRight.x += 0.5f * playerWidth;
        playerTopRight.y += playerWidth;
        playerTopRight = recanonicalisePosition(tilemap, playerTopRight);

        if (
          isTileMapPointEmpty(tilemap, playerTopLeft) &&
          isTileMapPointEmpty(tilemap, playerTopRight) &&
          isTileMapPointEmpty(tilemap, playerBottomLeft) &&
          isTileMapPointEmpty(tilemap, playerBottomRight)
        ) {
            state->playerPos = newPlayerPos;
        }
    }

    // clearscreen
    drawRectangle(videoBuf, 0.0f, 0.0f, (real32)videoBuf->width, (real32)videoBuf->height, 1.0f, 0.0f, 1.0f);

    for (int32 relRow = -10; relRow < 10; relRow++) {
        for (int32 relCol = -20; relCol < 20; relCol++) {
            uint32 col = state->playerPos.absTileX + relCol;
            uint32 row = state->playerPos.absTileY + relRow;
            int32 tileId = getTileValue(tilemap, col, row);
            real32 fill = 0.5f;
            if (tileId == 1) {
                fill = 1.0f;
            } 
            if ((row == state->playerPos.absTileY) && (col == state->playerPos.absTileX)) {
                fill = 0.0f;
            }
            real32 minX = screenCenterX - tilemap->metersToPixels*state->playerPos.x + ((real32)relCol)*tilemap->tileSideInPixels;
            real32 minY = screenCenterY + tilemap->metersToPixels*state->playerPos.y - ((real32)relRow)*tilemap->tileSideInPixels;
            real32 maxX = minX + tilemap->tileSideInPixels;
            real32 maxY = minY - tilemap->tileSideInPixels;
            drawRectangle(videoBuf, minX, maxY, maxX, minY, fill, fill, fill);
        }
    }

    real32 playerLeft = screenCenterX - 0.5f*tilemap->metersToPixels*playerWidth;
    real32 playerTop = screenCenterY - tilemap->metersToPixels*playerHeight;
    drawRectangle(videoBuf, playerLeft, playerTop, playerLeft + tilemap->metersToPixels*playerWidth, playerTop + tilemap->metersToPixels*playerHeight, playerR, playerG, playerB);
}

extern "C" GAME_GET_SOUND_SAMPLES(gameGetSoundSamples) {
    outputSound(soundBuf, (GameState*)memory->permanentStorage);
}
