#pragma once
#include "handmade.h"

inline TileChunk *getTileChunk(TileMap *tileMap, uint32 tileChunkX, uint32 tileChunkY) {
    TileChunk *tileChunk = NULL;

    if ((tileChunkX >= 0) && (tileChunkX < tileMap->tileChunkCountX) &&
        (tileChunkY >= 0) && (tileChunkY < tileMap->tileChunkCountY)
    ) {
        tileChunk = &tileMap->tileChunks[tileChunkY * tileMap->tileChunkCountX + tileChunkX];
    }

    return tileChunk;
}

inline void recanonicaliseOrd(TileMap *tileMap, uint32 *tileOrd, real32* ord) {
    int32 offset = floorReal32ToInt32(*ord / tileMap->tileSideInMeters);

    *tileOrd += offset;
    *ord -= offset*tileMap->tileSideInMeters;

    Assert(*ord >= 0.0f);
    Assert(*ord < tileMap->tileSideInMeters);
}

inline TileChunkPosition getChunkPosition(TileMap* tileMap, uint32 absTileX, uint32 absTileY) {
    TileChunkPosition result = {};
    result.tileChunkX = absTileX >> tileMap->chunkShift;
    result.tileChunkY = absTileY >> tileMap->chunkShift;
    result.relTileX = absTileX & tileMap->chunkMask; 
    result.relTileY = absTileY & tileMap->chunkMask; 
    return result;
}

inline uint32 getTileValueUnchecked(TileMap *tileMap, TileChunk *tileChunk, uint32 testTileX, uint32 testTileY) {
    Assert(tileChunk);
    Assert(testTileX < tileMap->chunkDim);
    Assert(testTileY < tileMap->chunkDim);
    return tileChunk->tiles[testTileY * tileMap->chunkDim + testTileX];
}

inline void setTileValueUnchecked(TileMap *tileMap, TileChunk *tileChunk, uint32 testTileX, uint32 testTileY, uint32 tileValue) {
    Assert(tileChunk);
    Assert(testTileX < tileMap->chunkDim);
    Assert(testTileY < tileMap->chunkDim);
    tileChunk->tiles[testTileY * tileMap->chunkDim + testTileX] = tileValue;
}

inline uint32 getTileValue(TileMap *tileMap, TileChunk* tileChunk, uint32 testTileX, uint32 testTileY) {
    uint32 tileValue = 0;
    if (tileChunk) {
        tileValue = getTileValueUnchecked(tileMap, tileChunk, testTileX, testTileY);
    }
    return tileValue;
}

inline void setTileValue(TileMap *tileMap, TileChunk* tileChunk, uint32 testTileX, uint32 testTileY, uint32 tileValue) {
    if (tileChunk) {
        setTileValueUnchecked(tileMap, tileChunk, testTileX, testTileY, tileValue);
    }
}

inline uint32 getTileValue(TileMap *tileMap, uint32 testTileX, uint32 testTileY) {
    uint32 tileValue = 0;
    TileChunkPosition chunkPos = getChunkPosition(tileMap, testTileX, testTileY);
    TileChunk *tileChunk = getTileChunk(tileMap, chunkPos.tileChunkX, chunkPos.tileChunkY);
    tileValue = getTileValue(tileMap, tileChunk, testTileX, testTileY);
    return tileValue;
}

inline TileMapPosition recanonicalisePosition(TileMap *tileMap, TileMapPosition pos) {
    TileMapPosition result = pos;

    recanonicaliseOrd(tileMap, &result.absTileX, &result.x);
    recanonicaliseOrd(tileMap, &result.absTileY, &result.y);

    return result;
}

internal bool32 isTileMapPointEmpty(TileMap *tileMap, TileMapPosition testPos) {
    bool32 isEmpty = false;

    TileChunkPosition chunkPos = getChunkPosition(tileMap, testPos.absTileX, testPos.absTileY);
    TileChunk *tileChunk = getTileChunk(tileMap, chunkPos.tileChunkX, chunkPos.tileChunkY);
    isEmpty = (getTileValue(tileMap, tileChunk, chunkPos.relTileX, chunkPos.relTileY) == 0);

    return isEmpty;
}

internal void setTileMapValue(MemoryArena *arena, TileMap *tileMap, uint32 absTileX, uint32 absTileY, uint32 tileValue) {
    TileChunkPosition chunkPos = getChunkPosition(tileMap, absTileX, absTileY);
    TileChunk *tileChunk = getTileChunk(tileMap, chunkPos.tileChunkX, chunkPos.tileChunkY);
    Assert(tileChunk);
    setTileValue(tileMap, tileChunk, chunkPos.relTileX, chunkPos.relTileY, tileValue);
}
